c----------------------------------------------------------------------
c --- METSCAN -- Surface Meteorological Data Quality Assurance Program
c----------------------------------------------------------------------
c
c --- METSCAN  Version: 4.0       Level: 010315                    MAIN
c
c     Copyright (c) 2001 by Exponent, Inc.
c
c --- Written by:  M. Fernau, V. Tino, E. Insley, J. Scire
c                  EARTH TECH/Sigma Research
c                  196 Baker Avenue
c                  Concord, MA  01742
c                  (508) 371-4200
c
c --- Updated 010315 from 960622 (D. Strimaitis)
c     - Y2K treatment of year enforced
c
c----------------------------------------------------------------------
c
      program metscan
c
c --- Version 010315;
c
c --- Check hourly surface meteorological data in CD144 format
c --- for invalid or suspicious values
c ---    allows specification of hours of operation
c
c --- inputs:
c
c                                                         Defaults
c       id      - station id (e.g., 99999)              --   *  <- No default
c       iyr     - year of met data (e.g., 1987)         --   *  <- No default
c       iexpmo  - month of first record                 --   1
c       iexpdy  - day  of first record                  --   1
c       iexphr  - hour of first record                  --   0
c       jwsmn   - minimum (non-calm) wind speed (knots) --   2
c                 (i.e., calm = ws=0, wd=0 is allowed)
c       jwsmx   - maximum wind speed (knots)            --  40
c       jtmx    - maximum temp. flag (deg. F)           -- 100
c       jtmin   - minimum temp. flag (deg. F)           --   0
c       jcmx    - maximum ceiling height (100s feet)    -- 350
c       jdelt   - maximum hourly temp. change (deg. F)  --  15
c       idelt   - maximum hourly RH change (%)          --  20
c       jrhold  - RH (%) for hour preceeding run        --  60
c       jtold   - temp. (deg. F) for hour preceeding    --   *  <- No default
c                 the first hour of the year (used to
c                 evaluate hourly temp. change for
c                 first hour of run)
c       mincc   - minimum sky cover for unlimited       --   3
c                 ceiling
c       ihrop(0:23) - hours of operation
c                 (0 = not operating, 1 = operating)    -- 24*1
c       jrhmin  - minimum relative humidity             --  10
c       jdat    - File format type (1=CD144,2=NCDC SAMSON)   * <- No Default
c----------------------------------------------------------------------
c
      include 'y2k.scn'
      common /a/ jd,jyr,jmo,jday,jhr,jceil,cdir,cspd,
     &           ctemp,crh,jcover
      integer idy(12),num(12),ihrop(0:23)
      character*1 alp(12),blank,jceil(3),jcover
      character*1 jcover1
      character*7 jceilc
      character*80 text
      character cdir*2,cspd*2,ctemp*3,crh*3
c
      data idy /31,28,31,30,31,30,31,31,30,31,30,31/
      data num /0,1,2,3,4,5,6,7,8,9,10,11/, ihrop/24*1/
      data alp /'0','1','2','3','4','5','6','7','8','9','-','X'/
      data blank /' '/
      data iunl /999/, jwsmn/2/, jwsmx/40/, jtmx/100/, jtmin/0/,
     &     jcmx/350/, jdelt/15/, iexphr/0/, iexpdy/1/, iexpmo/1/,
     &     mincc/3/, jrhmin/10/, idelt/20/, jrhold/60/
c
      namelist /opts/ id,iyr,iexpmo,iexpdy,iexphr,jwsmn,jwsmx,jtmx,
     &         jtmin,jdelt,jtold,mincc,jcmx,ihrop,jrhmin,idelt,jrhold,
     &         jdat
c
c --- open files
c
      open(15, file='metscan.inp', status='old')
      open(16, file='metscan.lst')
c
c --- read inputs (Namelist format)
c
      read(15,opts)

c --- Set century using starting year of data
      call QAYR4(16,iyr,0,ierr)
      if(ierr.NE.0) stop 'Use a 4-digit year for start of scan'

      write(16,101) id,iyr
 101  format(/' data checked for station: ',i5,2x,'year: ',i4//)
      write(16,opts)
      write(16,102)
 102  format(1x)
c
      if (jdat.eq.1) open(8, file='cd144.dat', status='old')
      if (jdat.eq.2) then
        open(8,file='NCDC_S.DAT',status='old')
        read(8,421,end=999)jd
        read(8,423) text
      endif
421   format(1x,i5)
423   format(a80)

c --- check for leap year
c
      if (mod(iyr,4) .eq. 0) idy(2) = 29
      iexphr = iexphr - 1
c
c --- keep track of hours for checking hours of operation
c
      ihr = -1
c
c --- calculate the next expected hour,day, and month
c
 1    continue
      ihr = ihr + 1
      if (ihr.eq.24) ihr = 0
      iexphr = iexphr+1
      if (iexphr .gt. 23) then
        iexphr = 0
        iexpdy = iexpdy + 1
        if (iexpdy .gt. idy(iexpmo)) then
          iexpdy = 1
          iexpmo = iexpmo + 1
          if (iexpmo .gt. 12) goto 999
        end if
      end if
c
c --- skip hour if station is not operating
      if (ihrop(ihr) .eq. 0) goto 1
c
c --- read in all data as character to compare to blanks
c
      if (jdat.eq.1) then
        read(8,103,end=999) jd,jyr,jmo,jday,jhr,jceil,cdir,cspd,
     &                      ctemp,crh,jcover
 103    format(i5,4i2,3a1,22x,2a2,4x,a3,3x,a3,23x,a1)

c ---   Change year to YYYY (Y2K)
        call YR4(16,jyr,ierr)
        if(ierr.NE.0) stop 'Halted in MAIN - see list file'
c
c ---   now convert cdir,cws,ctemp, & crh to integer for checking
c
        read(cdir,104) jdir
        read(cspd,104) jspd
 104    format(i2)
        read(ctemp,105) jtemp
        read(crh,105) jrh
 105    format(i3)
      endif
      if (jdat.eq.2) then
        read(8,422,end=999)jyr,jmo,jday,jhr,jcover1,jcover,temp,jrh,
     >                     jdir,spd,jceilb
c
c ---   Change year to YYYY (Y2K)
        call YR4(16,jyr,ierr)
        if(ierr.NE.0) stop 'Halted in MAIN - see list file'
c
c ---   Check for leap year
        if (mod(jyr,4) .eq. 0) idy(2) = 29
c
30      do ii=1,12
          if (jhr.eq.24.and.jmo.eq.ii.and.jday.eq.idy(ii)) then
            jhr=0
            jday=1
            jmo=ii+1
          endif
        enddo
        if (jmo.eq.13) then
          jhr=0
          jday=1
          jmo=1
          jyr=jyr+1
        endif
        if (jhr.eq.24) then
          jhr=0
          jday=jday+1
        endif
        if (jcover1.eq.'1') jcover='-'
        if (jcover1.eq.'9') jcover=' '
        if (temp.eq.9999.) then
          jtemp=999
        else
          jtemp=nint((9./5.*temp)+32.)
        endif
        if (jdir.eq.999) then
           jdir=99
        else
           jdir=jdir/10
        endif
        if (spd.eq.9999..or.spd.eq.99.) then
           jspd=99
        else
           jspd=nint(spd*1.94254)
        endif
        if (jceilb.eq.77777) then
          jceil(1)='-'
          jceil(2)='-'
          jceil(3)='-'
        elseif (jceilb.eq.88888) then
          jceil(1)='8'
          jceil(2)='8'
          jceil(3)='8'
        elseif (jceilb.eq.999999) then
          jceil(1)=' '
          jceil(2)=' '
          jceil(3)=' '
        else
          jceilb=nint(jceilb*3.2808399/100.)
          write(jceilc,425) jceilb
425       format(I7)
          if (jceilc(5:5).eq.' ') jceilc(5:5)='0'
          if (jceilc(6:6).eq.' ') jceilc(6:6)='0'
          if (jceilc(7:7).eq.' ') jceilc(7:7)='0'
          jceil(3)=jceilc(7:7)
          jceil(2)=jceilc(6:6)
          jceil(1)=jceilc(5:5)
        endif
      endif
422   format(1x,i2,3(1x,i2),40x,a1,a1,f6.1,6x,i4,5x,i4,f6.1,7x,i7)
c
c --- write day to screen to track progress of program
c
      if (jhr .eq. 0) write(*,*) 'jmo = ',jmo,' jday = ',jday
c
c --- check station id
c
      if (jd .ne. id) then
        write(16,106) id
 106    format(' id = ',i5,' does not match jd')
        call dump
      end if
c
c --- check for missing hours
c
      if (jyr .ne. iyr) then
        write(16,108) iyr
 108    format(' iyr = ',i4,' does not match jyr')
        call dump
        goto 1001
      end if
      if (jmo .ne. iexpmo) then
        write(16,109) iexpmo
 109    format(' iexpmo = ',i2,' does not match jmo')
        call dump
        goto 1001
      end if
      if (jday .ne. iexpdy) then
        write(16,110) iexpdy
 110    format(' iexpdy = ',i2,' does not match jday')
        call dump
        goto 1001
      end if
      if (jhr .ne. iexphr) then
        write(16,111) iexphr
 111    format(' iexphr = ',i2,' does not match jhr')
        call dump
        goto 1001
      end if
c
c --- check for missing data.  if all data for the hour is missing, just
c --- move on to the next record and check that.
c
      if (jceil(1) .eq. ' ' .and. jceil(2) .eq. ' ' .and. jceil(3)
     &    .eq. ' ' .and. cdir .eq. '  ' .and. cspd .eq. '  ' .and.
     &    ctemp .eq. '   ' .and. crh .eq. '   ' .and. jcover
     &    .eq. ' ') goto 1
c
c --- check relative humidity
c
c <MEF: many blanks in a row may throw off the jrhold check>
c
      if (crh .ne. '   ') then
        if (jrh .lt. jrhmin .or. jrh .gt. 100) then
          write (16,112) jrh
 112      format(' relative humidity flag --- jrh = ',i3)
          call dump
        end if
        if (iabs(jrh-jrhold) .ge. idelt) then
          write(16,113) jrh,jrhold,idelt
 113      format(' delta rh flag -- jrh = ',i3,' jrhold = ',i3,
     &           ' idelt = ',i3)
          call dump
        end if
        jrhold=jrh
      end if
c
c --- check wind speed and direction
c
c --- the following if-then will flag simultaneous missing ws and wd
c
      if (cspd .eq. '  ' .and. cdir .eq. '  ') then
        write(16,214)
 214    format(' missing ws and wd')
        call dump
      endif
      if (cspd .eq. '  ' .and. cdir .ne. '  ') then
        write(16,114) cspd,cdir
 114    format(' missing ws = ',a2,' nonmissing wd = ',a2)
        call dump
        if (jdir .lt.0 .or. jdir .gt. 36) then
          write(16,115) jdir
 115      format(' wind direction flag -- jdir = ',i3)
          call dump
        end if
      else if (cspd .ne. '  ' .and. cdir .eq. '  ') then
        write(16,116) cspd,cdir
 116    format(' nonmissing ws = ',a2,' missing wd = ',a2)
        call dump
        if (jspd .lt. jwsmn .or. jspd .gt. jwsmx) then
          write(16,117) jspd
 117      format(' wind speed flag -- jspd = ',i3)
          call dump
        end if
      else if (cspd .ne. '  ' .and. cdir .ne. '  ') then
c
c ----- the following if-then will skip checking calm winds
c
        if (jspd .eq. 0 .and. jdir .ne. 0) then
          write(16,118) jspd,jdir
 118      format(' zero ws = ',i2,' nonzero wd = ',i2)
          call dump
          if (jdir .lt.0 .or. jdir .gt. 36) then
            write(16,115) jdir
            call dump
          end if
        else if (jspd .ne. 0 .and. jdir .eq. 0) then
          write(16,119) jspd,jdir
 119      format(' nonzero ws = ',i2,' zero wd = ',i2)
          call dump
          if (jspd .lt. jwsmn .or. jspd .gt. jwsmx) then
            write(16,117) jspd
            call dump
          end if
        else if (jspd. ne .0 .and. jdir .ne. 0) then
          if (jdir .lt.0 .or. jdir .gt. 36) then
            write(16,115) jdir
            call dump
          end if
          if (jspd .lt. jwsmn .or. jspd .gt. jwsmx) then
            write(16,117) jspd
            call dump
          end if
        end if
      end if
c
c --- check temperature
c
c <MEF: many blanks in a row may throw off the jtold check>
c
      if (ctemp .ne. '   ') then
        if (jtemp .lt. jtmin .or. jtemp .gt. jtmx) then
          write(16,120) jtemp
 120      format(' temperature flag -- jtemp = ',i3)
          call dump
        end if
        if (iabs(jtemp-jtold) .ge. jdelt) then
          write(16,121) jtemp,jtold,jdelt
 121      format(' delta temp. flag -- jtemp = ',i3,' jtold = ',i3,
     &           ' jdelt = ',i3)
          call dump
        end if
        jtold=jtemp
      end if
c
c --- check opaque sky cover
c
      numc = 99
      if (jcover .ne. ' ') then
        do i=1,12
          if (jcover .eq. alp(i)) then
            numc = num(i)
            goto 2
          end if
        end do
        write(16,122) jcover
 122    format(' jcover is not one of the allowed characters -- ',
     &         'jcover = ',a1)
        call dump
 2      continue
      end if
c
c --- check ceiling height
c
      jbad = 0
      imiss = 0
      jmiss = 0
      junlim = 0
      do i = 1,12
        if (jceil(1) .eq. alp(i)) then
          if (i.lt.11) then
            nceil1 = num(i)
          else
            nceil1 = 9
            junlim = junlim + 1
          end if
          goto 3
        end if
      end do
      if (jceil(1) .eq. blank) then
        jmiss = jmiss + 1
        nceil1 = 0
      else
        jbad = jbad + 1
      end if
 3    continue
      do i = 1,12
        if (jceil(2) .eq. alp(i)) then
          if (i.lt.11) then
            nceil2 = num(i)
          else
            nceil2 = 9
            junlim = junlim + 1
          end if
          goto 4
        end if
      end do
      if (jceil(2) .eq. blank) then
        if (jmiss .ne. 1) imiss = imiss + 1
        jmiss = jmiss + 1
        nceil2 = 0
      else
        jbad = jbad + 1
      end if
 4    continue
      do i = 1,12
        if (jceil(3) .eq. alp(i)) then
          if (i.lt.11) then
            nceil3 = num(i)
          else
            nceil3 = 9
            junlim = junlim + 1
          end if
          goto 5
        end if
      end do
      if (jceil(3) .eq. blank) then
        if (jmiss .ne. 2) imiss = imiss + 1
        jmiss = jmiss + 1
        nceil3 = 0
      else
        jbad = jbad + 1
      end if
 5    continue
c
c --- the following if-then should skip missing ceilings
c
      if (junlim .gt. 0 .and. junlim .ne. 3) then
        write(16,123) jceil
 123    format(' problem with unlimited in ceiling = ',3a1)
        call dump
      else if (imiss .ne. 0) then
        write(16,124) jceil
 124    format(' problem with blanks in ceiling = ',3a1)
        call dump
      else if (jbad .ne. 0) then
        write(16,125) jceil
 125    format(' nonallowable characters in ceiling = ',3a1)
        call dump
      else if (jmiss .ne. 3) then
        nceil= nceil1 * 100 + nceil2 * 10 + nceil3
        if (numc .ne. 99) then
          if (numc .le. mincc .and. nceil .ne. iunl) then
            write(16,126) numc,nceil
 126        format(' cc = ',i2,' should not have ceiling = ',i5)
            call dump
          end if
        end if
        if (nceil .gt. jcmx .and. nceil .ne. iunl) then
          write(16,127) nceil
 127      format(' ceiling too high -- nceil = ',i5)
          call dump
        end if
      end if
c
c --- loop back for next hour
c
      go to 1
c
 1001 continue
      write(16,1002) iexphr,jhr,iexpdy,jday,iexpmo,jmo
 1002 format(' stop: missing record(s)? -- iexphr = ',i2,
     &       ' jhr = ',i2,' iexpdy = ',i2,' jday = ',i2,
     &       ' iexpmo = ',i2,' jmo = ',i2)
      call dump
 999  continue
      write(16,1003) jyr,jmo,jday,jhr
 1003 format(' last time period processed: jyr = ',i4,
     &       ' jmo = ',i2,' jday = ',i2,' jhr = ',i2//)
c
      stop
      end
c
c----------------------------------------------------------------------
      subroutine dump
c----------------------------------------------------------------------
c
c --- METSCAN  Version: 4.0       Level: 010315                    DUMP
c
c --- dump the record to the file
c----------------------------------------------------------------------
c
      common /a/ jd,jyr,jmo,jday,jhr,jceil,cdir,cspd,
     &           ctemp,crh,jcover
      character*1 jceil(3),jcover
      character cdir*2,cspd*2,ctemp*3,crh*3
      write(16,107) jd,jyr,jmo,jday,jhr,jceil,cdir,cspd,
     &              ctemp,crh,jcover
 107  format(i5,i4,3i2,' ceil ',3a1,' dir,spd ',2a2,' temp ',a3,
     &       ' rh ',a3,' cover ',a1)
      return
      end
c----------------------------------------------------------------------
      subroutine yr4(io,iyr,ierr)
c----------------------------------------------------------------------
c
c --- METSCAN  Version: 4.0       Level: 010315                     YR4
c ---          D. Strimaitis,   Earth Tech
c
c --- PURPOSE: Checks/converts 2-digit year to 4-digit year
c
c --- INPUTS:
c                IO - integer    - Unit number for list file output
c               IYR - integer    - Year (YYYY or YY)
c
c     Common block /Y2K/:
c           IYYLO, ICCLO
c
c --- OUTPUT:
c               IYR - integer    - Year (YYYY)
c              IERR - integer    - Error code: 0=OK, 1=FATAL
c
c --- YR4 called by:  Input routines reading 'year' data
c --- YR4 calls:      none
c----------------------------------------------------------------------
c
c --- Include commons
      include 'y2k.scn'

      ierr=0

c --- Test for 4-digit year (must exceed 1000)
      if(iyr.GT.1000) then
c ---    Passes 11th Century test (large year not trapped)
         return
      elseif(iyr.LT.100 .AND. iyr.GE.0) then
c ---    2-digit year
c ---    Construct 4-digit year
         if(iyr.LT.iyylo) then
            iyr=(icclo+1)*100+iyr
         else
            iyr=icclo*100+iyr
         endif
      else
c ---    Year not recognized
         ierr=1
         write(io,*)'ERROR in YR4 --- Year not recognized: ',iyr
         write(*,*)'ERROR in YR4 --- Year not recognized: ',iyr
      endif

      return
      end
c----------------------------------------------------------------------
      subroutine qayr4(io,iyr,metrun,ierr)
c----------------------------------------------------------------------
c
c --- METSCAN  Version: 4.0       Level: 010315                   QAYR4
c ---          D. Strimaitis,   Earth Tech
c
c --- PURPOSE: Defines century and year markers to use in converting
c ---          2-digit year to 4-digit year
c ---          The IBYR (YYYY) must be provided in the control file
c
c --- INPUTS:
c                IO - integer    - Unit number for list file output
c               IYR - integer    - Year provided for start of run
c            METRUN - integer    - Flag to run period in met file
c                                  0 = do not run period
c                                  1 = run period
c
c --- OUTPUT:
c              IERR - integer    - Error code: 0=OK, 1=FATAL
c
c     Common block /Y2K/:
c           IYYLO, ICCLO
c
c --- QAYR4 called by:  READCF
c --- QAYR4 calls:      none
c----------------------------------------------------------------------
c --- Include commons
      include 'y2k.scn'

c --- Sets parameters for the starting century marker (CC) and the
c --- 2-digit year (YY) used as the marker between the starting century
c --- and the next century.  For example, if CC=19 and YY=30, then a
c --- year less than 30 (say 15) is assumed to be 2015.  Any year
c --- greater than or equal to 30 (say 56) is assumed to be 1956.

c --- Set number of years prior to start of simulation that must not
c --- be placed in the next century
      data ibackyr/20/

      ierr=0

c --- Expect explicit starting year (YYYY)
c --- Test for 4-digit year (must exceed 1000)
      if(iyr.GT.1000) then
c ---    Passes 11th Century test (large year not trapped)
c ---    Back up IBACKYR years to set IYYLO
         kyr=iyr-ibackyr
c ---    Extract starting 2-digit century and 2-digit year
         icclo=kyr/100
         iyylo=kyr-icclo*100

c ---    Warn user that control file input is used to convert to YYYY
         iyr1=icclo*100+iyylo
         iyr2=(icclo+1)*100+iyylo-1
         write(io,*)
         write(io,*)
         write(io,*)
         write(io,*)'-------------------------------------------------'
         write(io,*)'NOTICE: Starting year in control file sets the'
         write(io,*)'        expected century for the simulation.  All'
         write(io,*)'        YY years are converted to YYYY years in'
         write(io,*)'        the range: ',iyr1,iyr2
         write(io,*)'-------------------------------------------------'
         write(io,*)
      else
         ierr=1
         write(*,*)
         write(*,*)'--------------------------------------------'
         write(*,*)'QAYR4 -- Start year must be 4-digits!: ',iyr
         if(metrun.EQ.1) then
            write(*,*)'         and must always be provided'
         endif
         write(*,*)'--------------------------------------------'
         write(*,*)
         write(io,*)
         write(io,*)'-------------------------------------------'
         write(io,*)'QAYR4 -- Start year must be 4-digits!: ',iyr
         if(metrun.EQ.1) then
            write(io,*)'         and must always be provided'
         endif
         write(io,*)'-------------------------------------------'
         write(io,*)
      endif

      return
      end
